#include<stdio.h>
#include<stdlib.h>
#include "llist.h"

int print_node_at(NODE *head, int pos){
//return:
// 0 : normal
// 1 : head is null
// 2 : pos is outside list
//prints by index
	NODE_INT currentNode;

	currentNode = traverse(head, pos);
	switch(currentNode.num){
		case 0:
			printf("The data at index %d is: %d\n", pos, currentNode.node->data);
			return 0;
		case 1:
			return 1;
		case 2:
			puts("That's outside the list!");
			return 2;
	}
}

void print_node(NODE *head){
//prints head
	printf("%d\n", head->data);}

void print_list(NODE *head){
	NODE *currentNode;
	currentNode = head;

	int i = 0;

	while (currentNode != NULL){
		printf("Index: %d; Data: %d\n",i,currentNode->data);
		currentNode=currentNode->next;
		i++;
	}
}

NODE_INT psearch(NODE *head, datatype data){
//search for a value. prints updates
//return:
// positive : index where data was found; NODE* return is pointer to where
// -1 : not found; NODE* return is NULL
	printf("Searching list for: %d\n", data);
	NODE_INT node = search(head, data);
	if(node.node != NULL)
		printf("Location found, index: %d.\n", node.num);
	else
		puts("Not found in this list.");
	return node;
}

NODE_INT search(NODE *head, datatype data){
//searches for a value.
//return:
// NODE* is NULL : not found; int is index of last null
// NODE* is non-null : int is index where data was found;
//---------------------NODE* return is pointer to where
	NODE_INT curr;
	curr.node = head;
	int i;
	for(i=0;curr.node != NULL;i++){
		if(curr.node->data == data){
			curr.num=i;
			return curr;
		}
		curr.node=curr.node->next;
	}
	curr.node = NULL;
	curr.num = i;
	return curr;
}

NODE_INT search_before(NODE *head, datatype data, unsigned before){
//searches for a value, then returns the item 'before' before it.
//only traverses the list once.
//return:
// node non-null, num>=0 : normal; found data at index[num+before]
// node NULL, num 0 : head is NULL
// node NULL, num<0 : found data before enough items processed. num is index.
//--------------------e.g. search_before({1,2,3},2,3) -> (NULL, -2)
// node NULL, num>0 : did not find data; returns index of final NULL
//--------------------e.g. search_before({1,2,3},4,3) -> (NULL, 3)
	NODE_INT out;
	out.node = NULL;
	out.num = 0;
	if(head==NULL)
		return out;
	if(before == 0){
		return search(head,data);
	}
	before;
	NODE **queue = calloc(before, sizeof(NODE*));
	int i=0;
	while(head != NULL && head->data != data){
		queue[i] = head;
		i = (i+1)%before;
		out.num++;
		head = head->next;
	}
	if(head == NULL){
		out.node = NULL;
		return out;
	}
	out.node = queue[i];
	out.num -= before;
	return out;
}

NODE *construct(datatype data, NODE *next){
//return:
// pointer to new node : normal
// NULL : malloc failed
	NODE *out = malloc(sizeof(NODE));
	if(out != NULL){
		out->data = data;
		out->next = next;
	}
	return out;
}

NODE_INT traverse(NODE *head, int pos){
//return:
// 0 : normal
// 1 : head is null; NODE* return is NULL
// 2 : reached end of list; NODE* return is final element
	NODE_INT out;
	out.node = head;
	out.num = 1;
	if(head==NULL)
		return out;
	int i;
	out.num = 0;
	for(i=0;i<pos || pos<0;i++){
		if(out.node->next != NULL)
			out.node = out.node->next;
		else{
			out.num = 2;
			break;
		}
	}
	return out;
}

NODE_INT traverse_before(NODE *head, int pos, unsigned before){
//return:
// 0 : normal
// 1 : head is null; NODE* return is NULL
// 2 : reached end of list; NODE* return is final item
// 3 : reached end of list, but before put result back in list: NODE* is result
// 4 : before is larger than pos or pos is negative; NODE* return is NULL
	int effectivePos = pos - before;
	NODE_INT out;
	out.node=NULL;
	out.num=4;
	if(effectivePos >= 0){
		out = traverse(head, effectivePos);
		if(out.num==0)
			out.num = length(out.node) > before ? 0 : 3;
	}
	return out;
}

int delete_node_after(NODE *head){
//return:
// 0 : normal
// 1 : head is NULL, or head->next is NULL, which means no node after to delete
	if(head==NULL || head->next == NULL)
		return 1;
	NODE* tmp = head->next;
	head->next = head->next->next;
	free(tmp);
	return 0;
}

void delete_list(NODE *head){
	if(head == NULL)
		return;
	NODE *tmp = head->next;
	free(head);
	delete_list(tmp);
}

int delete_node_at(NODE *head, int pos){
//return:
// 0 : normal
// 1 : head is NULL
// 2 : head->next is NULL: this function will not free head
// 3 : pos is beyond end of list
	if(head == NULL)
		return 1;
	if(head->next == NULL)
		return 2;
	if(pos == 0){
		head->data = head->next->data;
		delete_node_after(head);
		return 0;
	}
	NODE_INT priorNode = traverse_before(head, pos, 1);
	if(priorNode.num == 2 || priorNode.num == 3)
		return 3;
	NODE *tmp = priorNode.node->next;
	priorNode.node->next = tmp->next;
	free(tmp);
	return 0;
}

int delete_node(NODE *head, datatype data){
//return:
// 0 : normal
// 1 : head is NULL
// 2 : data is sole item in list: this function will not free head
// 4 : data not found
	NODE_INT result = search_before(head,data,1);
	if(result.node == NULL){
		if(result.num > 0)//not found
			return 4;
		if(result.num == 0)//head is null
			return 1;
		if(head->next == NULL)//data is sole item in list
			return 2;
		//data is first item: preserve address of head while deleting it
		head->data = head->next->data;
		delete_node_after(head);
		return 0;
	}
	//data is somewhere in the list
	delete_node_after(result.node);
	return 0;
}

NODE *construct_list(datatype list[], int size){
//return:
// pointer to head : normal
// NULL : malloc failed
	if(size<=0)
		return NULL;
	NODE *next = construct(list[size-1],NULL);
	NODE *head;
	int i;
	for(i=size-2;i>=0;i--){
		head = next;
		next = construct(list[i],head);
		if(next == NULL)
			delete_list(head);
	}
	return next;
}

int add_to_end(NODE *head, datatype new_data){
	return insert_node(head, -1, new_data);
}

int insert_node(NODE *head, int pos, datatype new_data){
//return:
// 0 : normal
// 1 : head is NULL
// 2 : malloc failed
// 3 : pos is larger than length(head)
	if(head==NULL)
		return 1;
	if(pos!=0){
		NODE_INT traverseResult = traverse(head,pos-1);
		int insertResult;
		if(traverseResult.num == 2 && pos > 0)
			return 3;
		return insert_node_after(traverseResult.node,new_data);
	}
	else{
		int out = insert_node_after(head,head->data);
		if(out>0)
			return out;
		head->data = new_data;
	}
	return 0;
}

int insert_node_after(NODE *head, datatype data){
//return:
// 0 : normal
// 1 : head is NULL
// 2 : malloc failed
	if(head==NULL)
		return 1;
	NODE *new = construct(data, head->next);
	if(new == NULL)
		return 2;
	head->next = new;
	return 0;
}


int length(NODE *head){
//returns length of list
	if(head==NULL)
		return 0;
	int i=1;
	while((head=head->next) != NULL)
		i++;
	return i;
}
